# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd
import tensorflow as tf

import causalimpact.inferences as inferrer
from causalimpact.misc import maybe_unstandardize


def test_get_lower_upper_percentiles():
    lower_percen, upper_percen = inferrer.get_lower_upper_percentiles(0.05)
    assert [lower_percen, upper_percen] == [2.5, 97.5]


def test_maybe_unstandardize():
    data = pd.DataFrame(np.arange(0, 10))
    results = inferrer.maybe_unstandardize(data)
    pd.testing.assert_frame_equal(results, data)

    mu_sig = (1, 2)
    results = inferrer.maybe_unstandardize(data, mu_sig)
    pd.testing.assert_frame_equal(results, data * 2 + 1)


def test_build_cum_index():
    data = pd.DataFrame(np.arange(0, 10))
    pre_data = data.iloc[:4]
    post_data = data.iloc[6:]
    new_index = inferrer.build_cum_index(pre_data.index, post_data.index)
    np.testing.assert_equal(new_index, np.array([3, 6, 7, 8, 9]))


def test_compile_posterior_inferences():
    data = pd.DataFrame(np.arange(10))
    pre_data = data.iloc[:3]
    post_data = data.iloc[7:]
    one_step_mean = 3
    one_step_stddev = 1.5
    posterior_mean = 7.5
    posterior_stddev = 1.5
    alpha = 0.05
    mu = 1
    sig = 2
    mu_sig = (mu, sig)
    niter = 10
    mask = np.array([True] * len(post_data))

    class OneStepDist:
        def sample(self, niter):
            tmp = tf.convert_to_tensor(
                np.tile(np.arange(start=3.1, stop=6.1, step=1), (niter, 1)) +
                np.arange(niter).reshape(-1, 1),
                dtype=np.float32
            )
            tmp = tmp[..., tf.newaxis]
            return tmp

        def mean(self):
            return np.ones((len(pre_data), 1)) * one_step_mean

        def stddev(self):
            return np.ones((len(pre_data), 1)) * one_step_stddev

    class PosteriorDist:
        def sample(self, niter):
            tmp = tf.convert_to_tensor(
                np.tile(np.arange(start=7.1, stop=10.1, step=1), (niter, 1)) +
                np.arange(niter).reshape(-1, 1),
                dtype=np.float32
            )
            tmp = tmp[..., tf.newaxis]
            return tmp

        def mean(self):
            return np.ones((len(post_data), 1)) * posterior_mean

        def stddev(self):
            return np.ones((len(post_data), 1)) * posterior_stddev

    one_step_dist = OneStepDist()
    posterior_dist = PosteriorDist()
    inferences = inferrer.compile_posterior_inferences(
        data.index,
        mask,
        pre_data,
        post_data,
        one_step_dist,
        posterior_dist,
        mu_sig,
        alpha=alpha,
        niter=niter
    )

    # by forcing the expected_index to be of type RangeIndex we avoid mismatching
    # types on Windows OS.
    expected_index = pd.RangeIndex(start=0, stop=6, step=1)

    # test complete_preds_means
    expec_complete_preds_means = pd.DataFrame(
        data=np.array([7, 7, 7, 16, 16, 16]),
        index=expected_index,
        dtype=np.float64,
        columns=['complete_preds_means']
    )
    pd.testing.assert_series_equal(
        expec_complete_preds_means['complete_preds_means'],
        inferences['complete_preds_means']
    )
    simulated_pre_ys = maybe_unstandardize(
        np.squeeze(one_step_dist.sample(niter)),
        mu_sig
    )
    simulated_post_ys = maybe_unstandardize(
        np.squeeze(posterior_dist.sample(niter)),
        mu_sig
    )
    lower_percen, upper_percen = inferrer.get_lower_upper_percentiles(alpha)
    # test complete_preds_lower
    pre_preds_lower, pre_preds_upper = np.percentile(
        simulated_pre_ys,
        [lower_percen, upper_percen],
        axis=0
    )
    post_preds_lower, post_preds_upper = np.percentile(
        simulated_post_ys,
        [lower_percen, upper_percen],
        axis=0
    )
    expec_complete_preds_lower = np.concatenate([pre_preds_lower, post_preds_lower])
    expec_complete_preds_lower = pd.DataFrame(
        data=expec_complete_preds_lower,
        index=expected_index,
        dtype=np.float64,
        columns=['complete_preds_lower']
    )
    pd.testing.assert_series_equal(
        expec_complete_preds_lower['complete_preds_lower'],
        inferences['complete_preds_lower']
    )
    # test complete_preds_upper
    expec_complete_preds_upper = np.concatenate([pre_preds_upper, post_preds_upper])
    expec_complete_preds_upper = pd.DataFrame(
        data=expec_complete_preds_upper,
        index=expected_index,
        dtype=np.float64,
        columns=['complete_preds_upper']
    )
    pd.testing.assert_series_equal(
        expec_complete_preds_upper['complete_preds_upper'],
        inferences['complete_preds_upper']
    )
    # test pre and post_preds_means
    expec_post_preds_means = pd.DataFrame(
        data=np.array([np.nan] * 3 + [posterior_mean * sig + mu] * len(pre_data)),
        index=expected_index,
        dtype=np.float64,
        columns=['post_preds_means']
    )
    pd.testing.assert_series_equal(
        expec_post_preds_means['post_preds_means'],
        inferences['post_preds_means']
    )
    # test post_preds_lower
    expec_post_preds_lower = pd.DataFrame(
        data=np.concatenate([[np.nan] * len(pre_data), post_preds_lower]),
        index=expected_index,
        dtype=np.float64,
        columns=['post_preds_lower']
    )
    pd.testing.assert_series_equal(
        expec_post_preds_lower['post_preds_lower'],
        inferences['post_preds_lower']
    )
    # test post_preds_upper
    expec_post_preds_upper = pd.DataFrame(
        data=np.concatenate([[np.nan] * len(pre_data), post_preds_upper]),
        index=expected_index,
        dtype=np.float64,
        columns=['post_preds_upper']
    )
    pd.testing.assert_series_equal(
        expec_post_preds_upper['post_preds_upper'],
        inferences['post_preds_upper']
    )
    # test post_cum_Y
    post_cum_y = np.concatenate([[np.nan] * (len(pre_data) - 1) + [0],
                                np.cumsum(post_data.iloc[:, 0])])
    expec_post_cum_y = pd.DataFrame(
        data=post_cum_y,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_y']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_y['post_cum_y'],
        inferences['post_cum_y']
    )
    # test post_cum_preds_means
    expec_post_cum_preds_means = np.cumsum(expec_post_preds_means)
    expec_post_cum_preds_means.rename(
        columns={'post_preds_means': 'post_cum_preds_means'},
        inplace=True
    )
    expec_post_cum_preds_means['post_cum_preds_means'][len(pre_data) - 1] = 0
    pd.testing.assert_series_equal(
        expec_post_cum_preds_means['post_cum_preds_means'],
        inferences['post_cum_preds_means']
    )
    # test post_cum_preds_lower
    post_cum_preds_lower, post_cum_preds_upper = np.percentile(
        np.cumsum(simulated_post_ys, axis=1),
        [100 * alpha / 2, 100 - 100 * alpha / 2],
        axis=0
    )
    post_cum_preds_lower = np.concatenate(
        [np.array([np.nan] * (len(pre_data) - 1) + [0]),
         post_cum_preds_lower]
    )
    expec_post_cum_preds_lower = pd.DataFrame(
        data=post_cum_preds_lower,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_preds_lower']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_preds_lower['post_cum_preds_lower'],
        inferences['post_cum_preds_lower']
    )
    # test post_cum_preds_upper
    post_cum_preds_upper = np.concatenate(
        [np.array([np.nan] * (len(pre_data) - 1) + [0]),
         post_cum_preds_upper]
    )
    expec_post_cum_preds_upper = pd.DataFrame(
        data=post_cum_preds_upper,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_preds_upper']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_preds_upper['post_cum_preds_upper'],
        inferences['post_cum_preds_upper']
    )
    # test point_effects_means
    net_data = pd.concat([pre_data, post_data])
    net_data.set_index(pd.RangeIndex(start=0, stop=6), inplace=True)
    expec_point_effects_means = net_data.iloc[:, 0] - inferences['complete_preds_means']
    expec_point_effects_means = pd.DataFrame(
        data=expec_point_effects_means,
        index=expected_index,
        dtype=np.float64,
        columns=['point_effects_means']
    )
    pd.testing.assert_series_equal(
        expec_point_effects_means['point_effects_means'],
        inferences['point_effects_means']
    )
    # test point_effects_lower
    expec_point_effects_lower = net_data.iloc[:, 0] - inferences['complete_preds_upper']
    expec_point_effects_lower = pd.DataFrame(
        data=expec_point_effects_lower,
        index=expected_index,
        dtype=np.float64,
        columns=['point_effects_lower']
    )
    pd.testing.assert_series_equal(
        expec_point_effects_lower['point_effects_lower'],
        inferences['point_effects_lower']
    )
    # test point_effects_upper
    expec_point_effects_upper = net_data.iloc[:, 0] - inferences['complete_preds_lower']
    expec_point_effects_upper = pd.DataFrame(
        data=expec_point_effects_upper,
        index=expected_index,
        dtype=np.float64,
        columns=['point_effects_upper']
    )
    pd.testing.assert_series_equal(
        expec_point_effects_upper['point_effects_upper'],
        inferences['point_effects_upper']
    )
    # test post_cum_effects_means
    post_data = post_data.set_index(pd.RangeIndex(start=3, stop=6))
    post_effects_means = post_data.iloc[:, 0] - inferences['post_preds_means']
    post_effects_means.iloc[len(pre_data) - 1] = 0
    expec_post_cum_effects_means = np.cumsum(post_effects_means)
    expec_post_cum_effects_means = pd.DataFrame(
        data=expec_post_cum_effects_means,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_effects_means']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_effects_means['post_cum_effects_means'],
        inferences['post_cum_effects_means']
    )
    # test post_cum_effects_lower
    post_cum_effects_lower, post_cum_effects_upper = np.percentile(
        np.cumsum(post_data.iloc[:, 0].values - simulated_post_ys, axis=1),
        [100 * alpha / 2, 100 - 100 * alpha / 2],
        axis=0
    )
    post_cum_effects_lower = np.concatenate(
        [np.array([np.nan] * (len(pre_data) - 1) + [0]),
         post_cum_effects_lower]
    )
    expec_post_cum_effects_lower = pd.DataFrame(
        data=post_cum_effects_lower,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_effects_lower']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_effects_lower['post_cum_effects_lower'],
        inferences['post_cum_effects_lower']
    )
    # test post_cum_effects_upper
    post_cum_effects_upper = np.concatenate(
        [np.array([np.nan] * (len(pre_data) - 1) + [0]),
         post_cum_effects_upper]
    )
    expec_post_cum_effects_upper = pd.DataFrame(
        data=post_cum_effects_upper,
        index=expected_index,
        dtype=np.float64,
        columns=['post_cum_effects_upper']
    )
    pd.testing.assert_series_equal(
        expec_post_cum_effects_upper['post_cum_effects_upper'],
        inferences['post_cum_effects_upper']
    )


def test_summarize_posterior_inference():
    post_preds_means = pd.Series([1, 2, 3])
    post_data = pd.DataFrame([1.1, 2.2, 3.3])
    simulated_ys = (
        np.tile(np.arange(start=7, stop=10, step=1), (10, 1)) +
        np.arange(10).reshape(-1, 1)
    )
    alpha = 0.05
    lower_percen, upper_percen = [100 * alpha / 2, 100 - 100 * alpha / 2]
    pred_lower, pred_upper = np.percentile(simulated_ys.mean(axis=1),
                                           [lower_percen, upper_percen])
    pred_cum_lower, pred_cum_upper = np.percentile(simulated_ys.sum(axis=1),
                                                   [lower_percen, upper_percen])

    summary = inferrer.summarize_posterior_inferences(post_preds_means, post_data,
                                                      simulated_ys, alpha)

    np.testing.assert_almost_equal(
        summary['average']['actual'],
        post_data.mean().values
    )
    np.testing.assert_almost_equal(
        summary['average']['predicted'],
        post_preds_means.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['predicted_lower'],
        pred_lower
    )
    np.testing.assert_almost_equal(
        summary['average']['predicted_upper'],
        pred_upper
    )
    np.testing.assert_almost_equal(
        summary['average']['abs_effect'],
        post_data.mean().values - post_preds_means.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['abs_effect_lower'],
        post_data.mean().values - pred_upper.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['abs_effect_upper'],
        post_data.mean().values - pred_lower.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['rel_effect'],
        (post_data.mean().values - post_preds_means.mean()) / post_preds_means.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['rel_effect_lower'],
        (post_data.mean().values - pred_upper.mean()) / post_preds_means.mean()
    )
    np.testing.assert_almost_equal(
        summary['average']['rel_effect_upper'],
        (post_data.mean().values - pred_lower.mean()) / post_preds_means.mean()
    )

    np.testing.assert_almost_equal(
        summary['cumulative']['actual'],
        post_data.values.sum()
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['predicted'],
        post_preds_means.sum()
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['predicted_lower'],
        pred_cum_lower
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['predicted_upper'],
        pred_cum_upper
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['abs_effect'],
        post_data.sum().values - post_preds_means.sum()
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['abs_effect_lower'],
        post_data.sum().values - pred_cum_upper
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['abs_effect_upper'],
        post_data.sum().values - pred_cum_lower
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['rel_effect'],
        (post_data.sum().values - post_preds_means.sum()) / post_preds_means.sum()
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['rel_effect_lower'],
        (post_data.sum().values - pred_cum_upper) / post_preds_means.sum()
    )
    np.testing.assert_almost_equal(
        summary['cumulative']['rel_effect_upper'],
        (post_data.sum().values - pred_cum_lower) / post_preds_means.sum()
    )


def test_compute_p_value():
    simulated_ys = (
        np.tile(np.arange(start=7, stop=10, step=1), (10, 1)) +
        np.arange(10).reshape(-1, 1)
    )
    post_data_sum = 46
    p_value = inferrer.compute_p_value(simulated_ys, post_data_sum)
    assert p_value == 2 / 11



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import numpy as np
import pandas as pd
import pytest
import tensorflow_probability as tfp

import causalimpact.data as cidata
import causalimpact.misc as misc


def test_format_input_data(rand_data):
    rand_data_32 = rand_data.astype(np.float32)
    data = cidata.format_input_data(rand_data)
    pd.testing.assert_frame_equal(data, rand_data_32)

    data = pd.DataFrame(cidata.format_input_data(rand_data.iloc[:, 0]))
    pd.testing.assert_frame_equal(data, pd.DataFrame(rand_data_32.iloc[:, 0]))

    with pytest.raises(ValueError) as excinfo:
        cidata.format_input_data('test')
    assert str(excinfo.value) == 'Could not transform input data to pandas DataFrame.'

    data = [1, 2, 3, 4, 5, 6, 2 + 1.j]
    with pytest.raises(ValueError) as excinfo:
        cidata.format_input_data(data)
    assert str(excinfo.value) == 'Input data must contain only numeric values.'

    data = np.random.randn(10, 2)
    data[0, 1] = np.nan
    with pytest.raises(ValueError) as excinfo:
        cidata.format_input_data(data)
    assert str(excinfo.value) == 'Input data cannot have NAN values.'


def test_process_pre_post_data(rand_data, date_rand_data):
    pre_data, post_data = cidata.process_pre_post_data(rand_data, [0, 10], [11, 20])
    freq_rand_data = rand_data.set_index(
        pd.date_range(start='2020-01-01', periods=len(rand_data))
    ).asfreq(pd.offsets.DateOffset(days=1))

    # Add +1 to make it inclusive on both ends
    pd.testing.assert_frame_equal(pre_data, freq_rand_data.iloc[0:10 + 1, :])
    pd.testing.assert_frame_equal(post_data, freq_rand_data.iloc[11:20 + 1, :])

    freq_date_rand_data = date_rand_data.asfreq(pd.offsets.DateOffset(days=1))
    pre_data, post_data = cidata.process_pre_post_data(
        freq_date_rand_data,
        ['20200101', '20200110'], ['20200111', '20200120']
    )
    pd.testing.assert_frame_equal(
        pre_data,
        freq_date_rand_data.loc['20200101': '20200110', :]
    )
    pd.testing.assert_frame_equal(
        post_data,
        freq_date_rand_data.loc['20200111': '20200120', :]
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [5, 10], [4, 7])
    assert str(excinfo.value) == (
        'Values in training data cannot be present in the '
        'post-intervention data. Please fix your pre_period value to cover at most one '
        'point less from when the intervention happened.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200101', '20200201'],
                                     ['20200110', '20200210'])
    assert str(excinfo.value) == (
        'Values in training data cannot be present in the '
        'post-intervention data. Please fix your pre_period value to cover at most one '
        'point less from when the intervention happened.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [5, 10], [15, 11])
    assert str(excinfo.value) == 'post_period last number must be bigger than its first.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [5, 10], [10, 15])
    assert str(excinfo.value) == ('post_period first value (10) must be bigger than the '
                                  'second value of pre_period (10).')

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200101', '20200110'],
                                     ['20200115', '20200111'])
    assert str(excinfo.value) == 'post_period last number must be bigger than its first.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [0, 2], [15, 11])
    assert str(excinfo.value) == 'pre_period must span at least 3 time points.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200101', '20200102'],
                                     ['20200115', '20200111'])
    assert str(excinfo.value) == 'pre_period must span at least 3 time points.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [5, 0], [15, 11])
    assert str(excinfo.value) == 'pre_period last number must be bigger than its first.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200105', '20200101'],
                                     ['20200115', '20200111'])
    assert str(excinfo.value) == 'pre_period last number must be bigger than its first.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200105', '20200110'],
                                     ['20200110', '20200115'])
    assert str(excinfo.value) == ('post_period first value (20200110) must be bigger '
                                  'than the second value of pre_period (20200110).')

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, 0, [15, 11])
    assert str(excinfo.value) == 'Input period must be of type list.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, '20200101',
                                     ['20200115', '20200130'])
    assert str(excinfo.value) == 'Input period must be of type list.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [0, 10, 30], [15, 11])
    assert str(excinfo.value) == (
        'Period must have two values regarding the beginning '
        'and end of the pre and post intervention data.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [0, None], [15, 11])
    assert str(excinfo.value) == 'Input period cannot have `None` values.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [0, 5.5], [15, 11])
    assert str(excinfo.value) == (
        'Input periods must contain either int, str or pandas Timestamp'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [-2, 10], [11, 20])
    assert str(excinfo.value) == '-2 not present in input data index.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, [0, 10], [11, 2000])
    assert str(excinfo.value) == '2000 not present in input data index.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(rand_data, ['20200101', '20200110'],
                                     ['20200111', '20200130'])
    assert str(excinfo.value) == (
        '20200101 not present in input data index.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20200101', '20200110'],
                                     ['20200111', '20220130'])
    assert str(excinfo.value) == ('20220130 not present in input data index.')

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(date_rand_data, ['20190101', '20200110'],
                                     ['20200111', '20200120'])
    assert str(excinfo.value) == ('20190101 not present in input data index.')

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(
            rand_data,
            [pd.Timestamp('20200101'), pd.Timestamp('20200110')],
            [pd.Timestamp('20200111'), pd.Timestamp('20200130')]
        )
    assert str(excinfo.value) == (
        '20200101 not present in input data index.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(
            date_rand_data,
            [pd.Timestamp('20200101'), pd.Timestamp('20200110')],
            [pd.Timestamp('20200111'), pd.Timestamp('20220130')]
        )
    assert str(excinfo.value) == ('20220130 not present in input data index.')

    with pytest.raises(ValueError) as excinfo:
        cidata.process_pre_post_data(
            date_rand_data,
            [pd.Timestamp('20190101'), pd.Timestamp('20200110')],
            [pd.Timestamp('20200111'), pd.Timestamp('20200120')]
        )
    assert str(excinfo.value) == ('20190101 not present in input data index.')


def test_validate_y():
    data = pd.Series(np.ones(100) * np.nan)
    with pytest.raises(ValueError) as excinfo:
        cidata.validate_y(data)
    assert str(excinfo.value) == 'Input response cannot have just Null values.'

    data[0:2] = 1
    with pytest.raises(ValueError) as excinfo:
        cidata.validate_y(data)
    assert str(excinfo.value) == ('Input response must have more than 3 non-null points '
                                  'at least.')

    data[0:3] = 1
    with pytest.raises(ValueError) as excinfo:
        cidata.validate_y(data)
    assert str(excinfo.value) == 'Input response cannot be constant.'


def test_convert_index_to_datetime():
    data = pd.DataFrame(np.random.rand(100, 2))
    assert isinstance(data.index, pd.core.indexes.range.RangeIndex)
    data = cidata.convert_index_to_datetime(data)
    assert isinstance(data.index, pd.core.indexes.range.RangeIndex)

    data.set_index(pd.date_range(start='20200101', periods=100), inplace=True)
    data.set_index(data.index.map(lambda x: str(x)[:-9]), inplace=True)
    assert isinstance(data.index.values[0], str)

    data = cidata.convert_index_to_datetime(data)
    assert isinstance(data.index, pd.core.indexes.datetimes.DatetimeIndex)

    data = pd.DataFrame([1, 2, 3])
    data.set_index(pd.Series(['A', 'B', 'C']), inplace=True)
    assert isinstance(data.index.values[0], str)
    data = cidata.convert_index_to_datetime(data)
    assert isinstance(data.index.values[0], str)


def test_convert_date_period_to_int(date_rand_data):
    result = cidata.convert_date_period_to_int(['20200101', '20200102'], date_rand_data)
    assert result == [0, 1]


def test_alpha_input():
    alpha = cidata.process_alpha(alpha=0.5)
    assert alpha == 0.5

    with pytest.raises(ValueError) as excinfo:
        cidata.process_alpha(alpha=1)
    assert str(excinfo.value) == 'alpha must be of type float.'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_alpha(alpha=2.)
    assert str(excinfo.value) == (
        'alpha must range between 0 (zero) and 1 (one) inclusive.'
    )

    with pytest.raises(ValueError) as excinfo:
        cidata.process_alpha(alpha=-0.5)
    assert str(excinfo.value) == (
        'alpha must range between 0 (zero) and 1 (one) inclusive.'
    )


def test_process_input_data(rand_data, pre_int_period, post_int_period, date_rand_data,
                            pre_str_period, post_str_period, monkeypatch):
    for type_ in ['int', 'str']:
        cur_data = rand_data if type_ == 'int' else date_rand_data
        cur_pre_period = pre_int_period if type_ == 'int' else pre_str_period
        cur_post_period = post_int_period if type_ == 'int' else post_str_period

        freq_data = cur_data.copy()
        if type_ == 'int':
            freq_data.set_index(
                pd.date_range(start='2020-01-01', periods=len(cur_data)),
                inplace=True
            )

        freq_data = freq_data.asfreq(pd.offsets.DateOffset(days=1))
        freq_data = freq_data.astype(np.float32)

        if type_ == 'int':
            cur_pre_data = freq_data.iloc[cur_pre_period[0]: cur_pre_period[1] + 1]
            cur_post_data = freq_data.iloc[cur_post_period[0]: cur_post_period[1] + 1]
        else:
            cur_pre_data = freq_data.loc[cur_pre_period[0]: cur_pre_period[1]]
            cur_post_data = freq_data.loc[cur_post_period[0]: cur_post_period[1]]

        results = cidata.process_input_data(cur_data, cur_pre_period, cur_post_period,
                                            None, {}, 0.05)

        pd.testing.assert_frame_equal(results['data'], cur_data.astype(np.float32))
        assert results['pre_period'] == cur_pre_period
        assert results['post_period'] == cur_post_period
        pd.testing.assert_frame_equal(
            results['pre_data'],
            cur_pre_data
        )
        pd.testing.assert_frame_equal(
            results['post_data'],
            cur_post_data
        )
        np.testing.assert_almost_equal(
            results['normed_pre_data'].mean().values,
            np.array([0, 0, 0])
        )
        np.testing.assert_almost_equal(
            results['normed_pre_data'].std().values,
            np.array([1, 1, 1]),
            decimal=2
        )
        assert isinstance(results['model'], tfp.sts.StructuralTimeSeries)
        assert results['model_args'] == {
            'standardize': True,
            'prior_level_sd': 0.01,
            'niter': 1000,
            'nseasons': 1,
            'season_duration': 1,
            'fit_method': 'vi'
        }
        assert results['alpha'] == 0.05

        pd.testing.assert_frame_equal(
            results['observed_time_series'],
            misc.standardize(pd.DataFrame(cur_pre_data.iloc[:, 0]))[0]
        )
        # tests user input model setting
        model = tfp.sts.LocalLevel()
        results = cidata.process_input_data(cur_data, cur_pre_period, cur_post_period,
                                            model, {}, 0.05)
        assert isinstance(results['model'], tfp.sts.LocalLevel)
        # test normalization of data not set
        results = cidata.process_input_data(cur_data, cur_pre_period, cur_post_period,
                                            None, {'standardize': False}, 0.05)
        assert results['normed_pre_data'] is None
        assert results['normed_post_data'] is None
    # test Exceptions
    with pytest.raises(ValueError) as excinfo:
        cidata.process_input_data(None, pre_int_period, post_int_period, None, {},
                                  0.05)
    assert str(excinfo.value) == 'data input argument cannot be empty'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_input_data(rand_data, None, post_int_period, None, {}, 0.05)
    assert str(excinfo.value) == 'pre_period input argument cannot be empty'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_input_data(rand_data, pre_int_period, None, None, {}, 0.05)
    assert str(excinfo.value) == 'post_period input argument cannot be empty'

    with pytest.raises(ValueError) as excinfo:
        cidata.process_input_data(None, None, post_int_period, None, {}, 0.05)
    assert str(excinfo.value) == 'data, pre_period input arguments cannot be empty'
    # testing calls
    format_input_data_mock = mock.Mock()
    format_input_data_mock.return_value = 'processed_data'
    process_pre_post_data_mock = mock.Mock()
    process_pre_post_data_mock.return_value = ['pre_data', 'post_data']
    process_alpha_mock = mock.Mock()
    process_alpha_mock.return_value = 0.05
    process_model_args_mock = mock.Mock()
    process_model_args_mock.return_value = {
        'standardize': True,
        'prior_level_sd': 0.01,
        'nseasons': 1,
        'season_duration': 1
    }
    check_input_model_mock = mock.Mock()
    build_default_model_mock = mock.Mock()
    build_default_model_mock.return_value = 'default model'
    standardize_pre_and_post_data_mock = mock.Mock()
    standardize_pre_and_post_data_mock.return_value = ('normed_pre_data',
                                                       'normed_post_data',
                                                       'mu_sig')
    build_observed_time_series_mock = mock.Mock()
    build_observed_time_series_mock.return_value = 'observed_time_series'
    build_mask_mock = mock.Mock(return_value='mask')

    monkeypatch.setattr('causalimpact.data.format_input_data', format_input_data_mock)
    monkeypatch.setattr('causalimpact.data.process_pre_post_data',
                        process_pre_post_data_mock)
    monkeypatch.setattr('causalimpact.data.process_alpha', process_alpha_mock)
    monkeypatch.setattr('causalimpact.data.cimodel.process_model_args',
                        process_model_args_mock)
    monkeypatch.setattr('causalimpact.data.cimodel.check_input_model',
                        check_input_model_mock)
    monkeypatch.setattr('causalimpact.data.cimodel.build_default_model',
                        build_default_model_mock)
    monkeypatch.setattr('causalimpact.data.standardize_pre_and_post_data',
                        standardize_pre_and_post_data_mock)
    monkeypatch.setattr('causalimpact.data._build_observed_time_series',
                        build_observed_time_series_mock)
    monkeypatch.setattr('causalimpact.data._build_nan_mask',
                        build_mask_mock)

    results = cidata.process_input_data('input_data', pre_int_period, post_int_period,
                                        'model', {}, 0.05)
    format_input_data_mock.assert_called_once_with('input_data')
    process_pre_post_data_mock.assert_called_once_with('processed_data', pre_int_period,
                                                       post_int_period)
    process_alpha_mock.assert_called_once_with(0.05)
    process_model_args_mock.assert_called_once_with({})
    check_input_model_mock.assert_called_once_with('model', 'pre_data', 'post_data')
    standardize_pre_and_post_data_mock.assert_called_once_with('pre_data', 'post_data')
    build_default_model_mock.assert_not_called()
    build_observed_time_series_mock.assert_called()

    assert results == {
        'data': 'processed_data',
        'pre_period': [0, 99],
        'post_period': [100, 199],
        'pre_data': 'pre_data',
        'post_data': 'post_data',
        'normed_pre_data': 'normed_pre_data',
        'normed_post_data': 'normed_post_data',
        'observed_time_series': 'observed_time_series',
        'model': 'model',
        'model_args': {
            'standardize': True,
            'prior_level_sd': 0.01,
            'nseasons': 1,
            'season_duration': 1
        },
        'alpha': 0.05,
        'mu_sig': 'mu_sig',
        'mask': 'mask'
    }

    results = cidata.process_input_data('input_data', pre_int_period, post_int_period,
                                        None, {}, 0.05)
    build_default_model_mock.assert_called_once_with('observed_time_series',
                                                     'normed_pre_data',
                                                     'normed_post_data', 0.01, 1, 1)
    assert results == {
        'data': 'processed_data',
        'pre_period': [0, 99],
        'post_period': [100, 199],
        'pre_data': 'pre_data',
        'post_data': 'post_data',
        'normed_pre_data': 'normed_pre_data',
        'normed_post_data': 'normed_post_data',
        'observed_time_series': 'observed_time_series',
        'model': 'default model',
        'model_args': {
            'standardize': True,
            'prior_level_sd': 0.01,
            'nseasons': 1,
            'season_duration': 1
        },
        'alpha': 0.05,
        'mu_sig': 'mu_sig',
        'mask': 'mask'
    }



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
import tensorflow_probability as tfp

import causalimpact.model as cimodel

tfd = tfp.distributions


def test_process_model_args():
    model_args = cimodel.process_model_args(dict(standardize=False))
    assert model_args['standardize'] is False

    model_args = cimodel.process_model_args(dict(standardize=True))
    assert model_args['standardize'] is True

    model_args = cimodel.process_model_args({})
    assert model_args['standardize'] is True

    with pytest.raises(ValueError) as excinfo:
        cimodel.process_model_args(dict(standardize='yes'))
    assert str(excinfo.value) == 'standardize argument must be of type bool.'

    model_args = cimodel.process_model_args(dict(niter=10))
    assert model_args['niter'] == 10

    model_args = cimodel.process_model_args({})
    assert model_args['niter'] == 1000

    with pytest.raises(ValueError) as excinfo:
        cimodel.process_model_args(dict(niter='yes'))
    assert str(excinfo.value) == 'niter argument must be of type int.'

    model_args = cimodel.process_model_args({})
    assert model_args['prior_level_sd'] == 0.01

    with pytest.raises(ValueError) as excinfo:
        cimodel.process_model_args(dict(prior_level_sd='test'))
    assert str(excinfo.value) == 'prior_level_sd argument must be of type float.'

    model_args = cimodel.process_model_args(dict(fit_method='hmc'))
    assert model_args['fit_method'] == 'hmc'

    model_args = cimodel.process_model_args(dict(fit_method='vi'))
    assert model_args['fit_method'] == 'vi'

    model_args = cimodel.process_model_args(dict())
    assert model_args['fit_method'] == 'vi'

    with pytest.raises(ValueError) as excinfo:
        model_args = cimodel.process_model_args(dict(fit_method='test'))
    assert str(excinfo.value) == 'fit_method can be either "hmc" or "vi".'

    model_args = cimodel.process_model_args(dict(nseasons=7))
    assert model_args['nseasons'] == 7

    model_args = cimodel.process_model_args({})
    assert model_args['nseasons'] == 1

    with pytest.raises(ValueError) as excinfo:
        model_args = cimodel.process_model_args(dict(nseasons='test'))
    assert str(excinfo.value) == 'nseasons argument must be of type int.'

    model_args = cimodel.process_model_args({})
    assert model_args['season_duration'] == 1

    model_args = cimodel.process_model_args(dict(nseasons=7, season_duration=24))
    assert model_args['season_duration'] == 24

    with pytest.raises(ValueError) as excinfo:
        model_args = cimodel.process_model_args(dict(season_duration='test'))
    assert str(excinfo.value) == 'season_duration argument must be of type int.'

    with pytest.raises(ValueError) as excinfo:
        model_args = cimodel.process_model_args(dict(season_duration=24))
    assert str(excinfo.value) == ('nseasons must be bigger than 1 when season_duration '
                                  'is also bigger than 1.')


def test_check_input_model():
    model = tfp.sts.Sum([tfp.sts.LocalLevel()])
    cimodel.check_input_model(model, None, None)

    model = tfp.sts.LocalLevel()
    cimodel.check_input_model(model, None, None)

    data = pd.DataFrame(np.random.rand(200, 2)).astype(np.float32)
    pre_data = data.iloc[:100, :]
    post_data = data.iloc[100:, :]
    model = tfp.sts.LinearRegression(design_matrix=data.iloc[:, 1].values.reshape(-1, 1))
    cimodel.check_input_model(model, pre_data, post_data)

    model = tfp.sts.LinearRegression(
        design_matrix=pre_data.iloc[:, 1].values.reshape(-1, 1)
    )
    with pytest.raises(ValueError) as excinfo:
        cimodel.check_input_model(model, pre_data, post_data)
    assert str(excinfo.value) == (
        'Customized Linear Regression Models must have total '
        'points equal to pre_data and post_data points and same number of covariates. '
        'Input design_matrix shape was (100, 1) and expected (200, 1) instead.'
    )

    model = tfp.sts.Sum([tfp.sts.LocalLevel(), tfp.sts.LinearRegression(
                        design_matrix=pre_data.iloc[:, 1].values.reshape(-1, 1))])
    with pytest.raises(ValueError) as excinfo:
        cimodel.check_input_model(model, pre_data, post_data)
    assert str(excinfo.value) == (
        'Customized Linear Regression Models must have total '
        'points equal to pre_data and post_data points and same number of covariates. '
        'Input design_matrix shape was (100, 1) and expected (200, 1) instead.'
    )
    with pytest.raises(ValueError) as excinfo:
        cimodel.check_input_model('test', None, None)
    assert str(excinfo.value) == 'Input model must be of type StructuralTimeSeries.'

    # tests dtype != float32
    data = pd.DataFrame(np.random.rand(200, 2))
    pre_data = data.iloc[:100, :]
    post_data = data.iloc[100:, :]
    model = tfp.sts.LinearRegression(design_matrix=data.iloc[:, 1].values.reshape(-1, 1))
    with pytest.raises(AssertionError):
        cimodel.check_input_model(model, pre_data, post_data)

    model = tfp.sts.LocalLevel(observed_time_series=pre_data.iloc[:, 0])
    with pytest.raises(AssertionError):
        cimodel.check_input_model(model, pre_data, post_data)

    model = tfp.sts.Sum(
        [tfp.sts.LinearRegression(design_matrix=data.iloc[:, 1].values.reshape(-1, 1)),
         tfp.sts.LocalLevel(observed_time_series=pre_data.iloc[:, 0])],
        observed_time_series=pre_data.iloc[:, 0]
    )
    with pytest.raises(AssertionError):
        cimodel.check_input_model(model, pre_data, post_data)


def test_build_default_model(rand_data, pre_int_period, post_int_period):
    prior_level_sd = 0.01
    pre_data = pd.DataFrame(rand_data.iloc[pre_int_period[0]: pre_int_period[1], 0])
    post_data = pd.DataFrame(rand_data.iloc[post_int_period[0]: post_int_period[1], 0])
    observed_time_series = pd.DataFrame(pre_data.iloc[:, 0]).astype(np.float32)
    model = cimodel.build_default_model(observed_time_series, pre_data, post_data,
                                        prior_level_sd, 1, 1)
    assert isinstance(model, tfp.sts.Sum)
    obs_prior = model.parameters[0].prior
    assert isinstance(obs_prior, tfd.TransformedDistribution)
    assert isinstance(obs_prior.bijector, tfp.bijectors.Power)
    assert isinstance(obs_prior.distribution, tfd.InverseGamma)
    level_prior = model.parameters[1].prior
    assert isinstance(level_prior, tfd.TransformedDistribution)
    assert isinstance(level_prior.bijector, tfp.bijectors.Power)
    assert isinstance(level_prior.distribution, tfd.InverseGamma)
    assert level_prior.dtype == tf.float32

    pre_data = pd.DataFrame(rand_data.iloc[pre_int_period[0]: pre_int_period[1], :])
    post_data = pd.DataFrame(rand_data.iloc[post_int_period[0]: post_int_period[1], :])
    observed_time_series = pd.DataFrame(pre_data.iloc[:, 0]).astype(np.float32)
    model = cimodel.build_default_model(observed_time_series, pre_data, post_data,
                                        prior_level_sd, 1, 1)
    assert isinstance(model, tfp.sts.Sum)
    obs_prior = model.parameters[0].prior
    assert isinstance(obs_prior, tfd.TransformedDistribution)
    assert isinstance(obs_prior.bijector, tfp.bijectors.Power)
    assert isinstance(obs_prior.distribution, tfd.InverseGamma)
    level_prior = model.parameters[1].prior
    assert isinstance(level_prior, tfd.TransformedDistribution)
    assert isinstance(level_prior.bijector, tfp.bijectors.Power)
    assert isinstance(level_prior.distribution, tfd.InverseGamma)
    assert level_prior.dtype == tf.float32
    linear = model.components[1]
    design_matrix = linear.design_matrix.to_dense()
    np.testing.assert_equal(pd.concat([pre_data, post_data]).iloc[:, 1:].values.astype(
                            np.float32),
                            design_matrix)
    assert design_matrix.dtype == tf.float32
    # test seasonal
    pre_data = pd.DataFrame(rand_data.iloc[pre_int_period[0]: pre_int_period[1], :])
    post_data = pd.DataFrame(rand_data.iloc[post_int_period[0]: post_int_period[1], :])
    observed_time_series = pd.DataFrame(pre_data.iloc[:, 0]).astype(np.float32)
    model = cimodel.build_default_model(observed_time_series, pre_data, post_data,
                                        prior_level_sd, 7, 2)
    assert isinstance(model, tfp.sts.Sum)
    obs_prior = model.parameters[0].prior
    assert isinstance(obs_prior, tfd.TransformedDistribution)
    assert isinstance(obs_prior.bijector, tfp.bijectors.Power)
    assert isinstance(obs_prior.distribution, tfd.InverseGamma)
    assert obs_prior.dtype == tf.float32
    level_prior = model.parameters[1].prior
    assert isinstance(level_prior, tfd.TransformedDistribution)
    assert isinstance(level_prior.bijector, tfp.bijectors.Power)
    assert isinstance(level_prior.distribution, tfd.InverseGamma)
    assert level_prior.dtype == tf.float32
    linear = model.components[1]
    design_matrix = linear.design_matrix.to_dense()
    np.testing.assert_equal(pd.concat([pre_data, post_data]).iloc[:, 1:].values.astype(
                            np.float32),
                            design_matrix)
    seasonal_component = model.components[-1]
    assert isinstance(seasonal_component, tfp.sts.Seasonal)
    assert seasonal_component.num_seasons == 7
    assert seasonal_component.num_steps_per_season == 2


def test_fit_model(monkeypatch):
    observed_time_series = 'observed_time_series'
    fit_mock = mock.Mock(return_value=('samples', 'kr'))
    vi_fit_mock = mock.Mock()

    class VarPost:
        def sample(self, *args):
            return f'{args[0]} var_post_samples'
    var_post = VarPost()
    surrogate_posterior_mock = mock.Mock(return_value=var_post)
    monkeypatch.setattr('causalimpact.model.tfp.sts.fit_with_hmc', fit_mock)
    monkeypatch.setattr('causalimpact.model.tfp.sts.build_factored_surrogate_posterior',
                        surrogate_posterior_mock)
    monkeypatch.setattr('causalimpact.model.Adam',
                        mock.Mock(return_value='optimizer'))

    monkeypatch.setattr('causalimpact.model.tfp.vi.fit_surrogate_posterior', vi_fit_mock)

    class Model:
        def joint_log_prob(self, observed_time_series):
            return f'target_log_prob of {observed_time_series}'
    model = Model()
    samples, kr = cimodel.fit_model(
        model=model,
        observed_time_series=observed_time_series,
        method='hmc'
    )
    fit_mock.assert_called_once_with(
        model=model, observed_time_series=observed_time_series
    )
    assert samples == 'samples'
    assert kr == 'kr'
    surrogate_posterior_mock.assert_not_called()

    samples, kr = cimodel.fit_model(
        model=model,
        observed_time_series=observed_time_series,
        method='vi'
    )
    surrogate_posterior_mock.assert_called_with(model=model)
    vi_fit_mock.assert_called_once_with(
        target_log_prob_fn='target_log_prob of observed_time_series',
        surrogate_posterior=var_post,
        optimizer='optimizer',
        num_steps=200
    )
    assert kr is None
    assert samples == '100 var_post_samples'

    with pytest.raises(ValueError) as excinfo:
        cimodel.fit_model(model=model, observed_time_series=observed_time_series,
                          method='test')
    assert str(excinfo.value) == (
        'Input method "test" not valid. Choose between "hmc" or "vi".'
    )
